;	=========================================
;	Wand-Funkuhr mit AT90S2313
;	=========================================
;	B0: Output LED - Segment a
;	B1: Output LED - Segment b
;	B2: Output LED - Segment c
;	B3: Output LED - Segment d
;	B4: Output LED - Segment e
;	B5: Output LED - Segment f
;	B6: Output LED - Segment g
;	B7: Output LED - Segment dp
;	-------------------------------
;	D0: Output 1 of 10 Decoder a
;	D1: Output 1 of 10 Decoder b
;	D2: Output 1 of 10 Decoder c
;	D3: Output 1 of 10 Decoder d
;	D4: Input DCF-Receiver
;	D5: Input Alarm
;	D6: Input free
;
;	-------------------------------
;	Timers
;	-------------------------------
;	Timer1: Interrupt every 1ms (Prescaler 1, Compare 4000)

;	----------------------------------
;	Standard definitions for AT90S2313
;	----------------------------------
.include "2313def.inc"

;	-------------------------------
;	SRAM
;	-------------------------------
.dseg
.org	0x60

disbeg:	.byte	16		;time data
alabeg:	.byte	16		;alarm data
dcfbeg:	.byte	64		;DCF data
;	-------------------------------
;	Constant definitions
;	-------------------------------
.equ	rambeg	=0x60		;RAM start
.equ	t1end	=4000		;Timer1 end value 1ms
.equ	t	=6		;t-flag position in SREG
.equ	dp	=7		;decimal point position
.equ	dispm	=10		;number of 7 segment displays
.equ	dcfsig	=4		;input DCF receiver
.equ	alasig	=5		;input alert mode
.equ	dcfzero	=6		;*10 ms DCF logical zero time
.equ	dcfone	=16		;*10 ms DCF logical one time
.equ	dcfmark	=160		;*10 ms DCF mark for new time frame
.equ	timeout	=250		;*10 ms DCF timeout (no Signal)
.equ	waiting	=250		;*10 ms waiting time for logo
.equ	dcfimp	=59		;max DCF pulses (59)
;	-------------------------------
;	Register definitions
;	-------------------------------
.def	stor	=r0		;storage variable for LPM
.def	flags	=r1		;flag storage
.def	stat	=r2		;status of parity
.def	temp	=r16		;temporary storage variable
.def	temp2	=r17		;temporary storage variable 2
.def	tcount	=r18		;temporary counter variable
.def	titemp	=r19		;timer ISR temporary storage variable
.def	dcfloop	=r20		;loop for DCF-Storage
.def	disco	=r21		;display counter
.def	tico10	=r22		;time counter increment every 10 ms
.def	tparty	=r23		;parity check

;	---------------------------------
;	Program start, Interrupt vectors
;	---------------------------------
.cseg
.org	0x0000

reset:	rjmp	initia			;program start
exint0:	reti
exint1:	reti
t1icap:	reti
t1coma:	rjmp	timer			;time counter
t1ovfl:	reti
t0ovfl:	reti
uartrx:	reti
uartem:	reti
uarttx:	reti
ancomp:	reti

;Initialize RAM, I/O and interrupts
initia:	cli				;disable interrupts

	ser	temp
	clr	temp2
	out	ddrb, temp
	out	portb,temp2		;set outputs
	ldi	temp, 0x0f
	out	ddrd, temp		;lower nibble outputs
	ldi	temp, 0x70
	out	portd,temp		;set pull-ups

	ldi	temp,ramend		;set stack pointer
	out	spl,temp

	ldi	yl,low(rambeg)		;RAM start low
	ldi	yh,high(rambeg)		;RAM start high
init:	st	y+,temp2		;clear RAM
	cpi	yl,ramend+1		;RAM end low?
	brne	init			;no -> loop

	ldi	temp,high(t1end)	;Timer1 end value a H
	out	ocr1ah,temp
	ldi	temp,low(t1end)		;Timer1 end value a L
	out	ocr1al,temp
	out	tccr1a,temp2		;Timer1 no PWM, no Compare-Out
	ldi	temp,(1<<ctc1)+(1<<cs10);Timer1 clear afer compare, prescaler = 1
	out	tccr1b,temp
	out	gimsk,temp2		;disable external interrupts
	ldi	temp,1<<ocie1a		;Timer1 compare1
	out	timsk,temp
	sei				;enable interrupts

	rcall	alert			;load Alarm text
	rcall	first			;show logo
	clr	tico10
wait:	cpi	tico10,waiting		;wait
	brlo	wait

;Main program
main:	clr	dcfloop			;jump here every minute

;duration high signal
start:	clr	tico10			;jump here every second
ifhigh:	cpi	tico10,timeout
	brlo	ifhi2
	rjmp	erdcfh			;broken wire
ifhi2:	brts	ifhigh			;wait while DCF (t-flag) high
	cpi	tico10,dcfmark
	brsh	main			;minute mark, no error

;duration low signal
	clr	tico10
iflow:	cpi	tico10,timeout
	brlo	iflo2
	rjmp	erdcfl			;short circuit
iflo2:	brtc	iflow			;wait while DCF (t-flag) low
	cpi	tico10,dcfone
	brsh	dcf1			;logical one
	cpi	tico10,dcfzero
	brsh	dcf0			;logical zero
	rjmp	ersync			;spike pulse error

dcf0:	clr	temp
	rjmp	main2
dcf1:	ldi	temp,1

main2:	ldi	zl,low(dcfbeg)
	ldi	zh,high(dcfbeg)
	add	zl,dcfloop
	st	z,temp			;store DCF pulses
	inc	dcfloop			;increment dcfloop
	cpi	dcfloop,dcfimp		;for all 59 pulses
	brlo	start

main3:	clr	stat			;check parity
	ldi	temp,21
	ldi	temp2,8
	rcall	parity
	tst	stat
	brne	errpar
	ldi	temp,29
	ldi	temp2,7
	rcall	parity
	tst	stat
	brne	errpar
	ldi	temp,36
	ldi	temp2,23
	rcall	parity
	tst	stat
	breq	main4

errpar:	lds	temp,disbeg		;parity error => all dots on
	sbr	temp,1<<dp
	sts	disbeg,temp
	lds	temp,disbeg+4
	sbr	temp,1<<dp
	sts	disbeg+4,temp
	lds	temp,disbeg+9
	sbr	temp,1<<dp
	sts	disbeg+9,temp
	rjmp	main

main4:	ldi	zh,high(dcfbeg)		;combine bits to bytes
	ldi	zl,low(dcfbeg)
	adiw	zl,25			;minute
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	sts	disbeg+0,temp		;minute ones

	adiw	zl,28
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	ori	temp,1<<dp		;point on
	sts	disbeg+1,temp		;minute tens

	adiw	zl,33			;hour
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	ori	temp,1<<dp		;point on
	sts	disbeg+2,temp		;hour ones

	adiw	zl,35
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	sts	disbeg+3,temp		;hour tens

	adiw	zl,40			;day
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	ori	temp,1<<dp		;point on
	sts	disbeg+8,temp		;day ones

	adiw	zl,42
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	sts	disbeg+9,temp		;day tens

	adiw	zl,49			;month
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	ori	temp,1<<dp		;point on
	sts	disbeg+6,temp		;month ones

	adiw	zl,49
	ld	temp,z
	rcall	fmread
	sts	disbeg+7,temp		;month tens

	adiw	zl,54			;year
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	sts	disbeg+4,temp		;year ones

	adiw	zl,58
	clr	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rol	temp
	ld	temp2,-z
	add	temp,temp2
	rcall	fmread
	sts	disbeg+5,temp		;year tens
	rjmp	main

;Timer1 ISR: use disco, titemp and pointer y only here
timer:	in	flags,sreg		;save flags
	in	titemp,pind		;store DCF status
	bst	titemp,dcfsig		;in T-Flag
	bld	flags,t
	ldi	yh,high(disbeg)
	ldi	yl,low(disbeg)+3	;Pointer on display RAM, hour tens
	ld	titemp,y
	bld	titemp,dp		;point shows DCF status
	st	y,titemp
	inc	disco			;increment display counter
	cpi	disco,dispm		;while disco < DispMax
	brne	go1
	inc	tico10			;increment time counter
	clr	disco			;clear disco
go1:	mov	yl,disco
	sbis	pind,alasig		;alert input
	rjmp	go2
	ldi	titemp,alabeg		;Pointer on alert RAM
	rjmp	go3
go2:	ldi	titemp,disbeg		;Pointer on time RAM
go3:	add	yl,titemp
	ldi	titemp,0xf0		;upper nibble pull-ups
	or	titemp,disco
	out	portd,titemp
	ld	titemp,y
	out 	portb,titemp
	out	sreg,flags		;restore flags
	reti

;Parity check
;temp  : in DCF offset
;temp2 : in number of bits
;stat  : out status
;Intern: tcount,tparty,z-pointer
parity:	clr	tparty
	clr	tcount
	ldi	zh,high(dcfbeg)
	ldi	zl,low(dcfbeg)
	add	zl,temp
pari2:	ld	temp,z+			;add bits
	add	tparty,temp
	inc	tcount
	cp	tcount,temp2
	brlo	pari2
	sbrc	tparty,0		;even?
	inc	stat
	ret

;FLASH Memory read
;temp  : in ASCII character
;temp  : out 16 segment data
;Intern: temp2,stor,z-pointer
fmread:	ldi	zh,high(segmtb*2)
	ldi	zl,low(segmtb*2)
	clr	temp2
	add	zl,temp
	adc	zh,temp2
	lpm
	mov	temp,stor
	ldi	zh,high(dcfbeg)
	ldi	zl,low(dcfbeg)
	ret

;Write a line on display
;x-pointer : write RAM (disbeg or alabeg)
;z-pointer : read FLASH
;Intern: temp,tcount,stor,stat,tparty
writel:	ldi	tcount,3
	rcall	write2
	ldi	tcount,9
	rcall	write2
	ret

write2:	add	xl,tcount
	lpm
	adiw	zl,1
	mov	temp,stor
	mov	stat,zl
	mov	tparty,zh
	rcall	fmread
	mov	zl,stat
	mov	zh,tparty
	st	x,temp
	sub	xl,tcount
	cpi	tcount,0
	breq	write3
	cpi	tcount,4
	breq	write3
	dec	tcount
	rjmp	write2
write3:	ret

; ----- Messages -----

erdcfl: ldi	xl,low(disbeg)		;DCF low (short circuit)
	ldi	xh,high(disbeg)
	ldi	zl,low(txdcf0*2)
	ldi	zh,high(txdcf0*2)
	rcall	writel
	rjmp	main

erdcfh: ldi	xl,low(disbeg)		;DCF high (broken wire)
	ldi	xh,high(disbeg)
	ldi	zl,low(txdcf1*2)
	ldi	zh,high(txdcf1*2)
	rcall	writel
	rjmp	main

ersync:	ldi	xl,low(disbeg)		;No sync
	ldi	xh,high(disbeg)
	ldi	zl,low(txsync*2)
	ldi	zh,high(txsync*2)
	rcall	writel
	rjmp	main

first:	ldi	xl,low(disbeg)		;Logo
	ldi	xh,high(disbeg)
	ldi	zl,low(txlogo*2)
	ldi	zh,high(txlogo*2)
	rcall	writel
	lds	temp,disbeg		;points on
	sbr	temp,1<<dp
	sts	disbeg,temp
	lds	temp,disbeg+6
	sbr	temp,1<<dp
	sts	disbeg+6,temp
	lds	temp,disbeg+8
	sbr	temp,1<<dp
	sts	disbeg+8,temp
	lds	temp,disbeg+9
	sbr	temp,1<<dp
	sts	disbeg+9,temp
	ret

alert:	ldi	xl,low(alabeg)		;Alarm
	ldi	xh,high(alabeg)
	ldi	zl,low(txalar*2)
	ldi	zh,high(txalar*2)
	rcall	writel
	ret

txdcf0:	.db	0x0d,0x0c,0x0f,0x1b,0x13,0x15,0x0e,0x17,0x17,0x1d	;dCF-LoErr!
txdcf1:	.db	0x0d,0x0c,0x0f,0x1b,0x10,0x11,0x0e,0x17,0x17,0x1d	;dCF-HiErr!
txsync:	.db	0x1e,0x14,0x15,0x1e,0x1e,0x05,0x1a,0x14,0x0c,0x1d	;_no__SYnC!
txlogo:	.db	0x00,0x13,0x0a,0x0f,0x02,0x02,0x00,0x03,0x07,0x07	;OLAF220377
txalar:	.db	0x17,0x0e,0x0d,0x1e,0x0a,0x13,0x0e,0x17,0x18,0x1d	;rEd_ALErt!

;		  pgfedcba,  pgfedcba
segmtb:	.db	0b00111111,0b00000110	;Code 00/01: Characters 0,1
	.db	0b01011011,0b01001111	;Code 02/03: Characters 2,3
	.db	0b01100110,0b01101101	;Code 04/05: Characters 4,5
	.db	0b01111101,0b00000111	;Code 06/07: Characters 6,7
	.db	0b01111111,0b01101111	;Code 08/09: Characters 8,9
	.db	0b01110111,0b01111100	;Code 0A/0B: Characters A,b
	.db	0b00111001,0b01011110	;Code 0C/0D: Characters C,d
	.db	0b01111001,0b01110001	;Code 0E/0F: Characters E,F
	.db	0b01110110,0b00010000	;Code 10/11: Characters H,i
	.db	0b00011110,0b00111000	;Code 12/13: Characters J,L
	.db	0b01010100,0b01011100	;Code 14/15: Characters n,o
	.db	0b01110011,0b01010000	;Code 16/17: Characters P,r
	.db	0b01111000,0b00111110	;Code 18/19: Characters t,U
	.db	0b01101110,0b01000000	;Code 1A/1B: Characters Y,-
	.db	0b01001000,0b10000010	;Code 1C/1D: Characters =,!
	.db	0b00001000,0b00000010	;Code 1E/1F: Characters _,'